%{/* -*- C -*- */
/* TOM parser generator lexer.
   Written by Pieter J. Schoenmakers <tiggr@ics.ele.tue.nl>

   Copyright (C) 1996-1998 Pieter J. Schoenmakers.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA

   $Id: gplex.l,v 1.21 1998/11/10 17:28:51 tiggr Exp $  */

#include "gp-r.h"
#include <ctype.h>
#include <float.h>
#include <limits.h>
#include <stdarg.h>
#include <tom/util.h>

/* limits.h on Debian 1.3 is broken.  */
#ifndef INT_MAX
#define INT_MAX ((int) (1LL << (8 * sizeof (int)) - 1))
#endif

#define YY_DECL  tom_int i_gp_GPLex_i_nextToken (tom_object self, selector cmd)

#define YY_USER_ACTION  \
  {								\
    struct _es_i_gps_SkeletonLexer *ilx				\
      = trt_ext_address (self, _ei_i_gps_SkeletonLexer);	\
    ilx->semantics = 0;

#define YY_BREAK  \
    break;	\
  }

#define yywrap()  1

#define YY_NO_UNPUT

#define YY_INPUT(buf,result,max_size) \
  do								\
    { 								\
      int c, n;							\
      for (n = 0;						\
	   (n < max_size					\
	    && (c = TRT_SEND (, c_gp_GPLex_lex_stream,		\
			      USEL (tom, i_read))) != -1);	\
	   ++n)							\
	buf[n] = (char) c; 					\
      result = n;						\
    } while (0)

void
error (char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  fputc ('\n', stderr);
  va_end (ap);
}

void
warning (int line, char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  if (!line)
    fprintf (stderr, "warning: ");
  else
    fprintf (stderr, "%d:warning: ", line);
  vfprintf (stderr, fmt, ap);
  fputc ('\n', stderr);
  va_end (ap);
}

tom_object
lex_string_cst (struct _es_i_gps_SkeletonLexer *ilx,
		const char *yytext, int yyleng, char quote)
{
  tom_object str = TRT_SEND ((reference_imp), _mr_c_tom_MutableByteString,
			     USEL (tom, r_withCapacity_i),
			     (tom_int) yyleng - 2);
  tom_int len;
  char *base;
  const char *s;
  int i, j;
  char c;

  TRT_SEND (, str, USEL (tom, v_resize_i), yyleng - 2);
  base = TRT_SEND ((pointer_imp), str, USEL (tom, _pi__byteStringContents),
		   &len);

  for (s = yytext + 1, j = 0, i = 1; i < yyleng - 1; s++, i++)
    {
      if (*s != '\\')
	c = *s;
      else
	{
	  i++;
	  if (*++s == quote)
	    c = quote;
	  else switch (*s)
	    {
	    case 'f': c = '\f'; break;
	    case 'n': c = '\n'; break;
	    case 'r': c = '\r'; break;
	    case 't': c = '\t'; break;
	    case '\\': c = '\\'; break;
	    default:
	      warning (ilx->current_line, "unknown escape: `%c'", *s);
	      base[j++] = '\\';
	      c = *s;
	      break;
	    }
	}
      base[j++] = c;
    }

  if (j != len)
    TRT_SEND (, str, USEL (tom, v_truncate_i), j);

  return str;
}

static void read_compound (struct _es_i_gps_SkeletonLexer *ilx,
			   char open, char close);
static void read_flat_compound (struct _es_i_gps_SkeletonLexer *ilx,
				char *close);

%}

%%

"//".*\n	{
		  struct _es_i_gp_GPLex *igp
		    = trt_ext_address (self, _ei_i_gp_GPLex);
		  if (igp->slash_slash_line != ilx->current_line - 1)
		    warning (ilx->current_line, "// comment");
		  igp->slash_slash_line = ilx->current_line++;
		}

"/*"		{
		  tom_int start_line = ilx->current_line;
		  int b = 0, c = input ();
		
		  while (c != EOF && !(c == '/' && b == '*'))
		    {
		      if (c == '\n')
			ilx->current_line++;
		      if (b == '/' && c == '*')
			warning (ilx->current_line, "`/*' in comment");
		      b = c;
		      c = input ();
		    }
		  if (c == EOF)
		    error ("%d: unterminated comment", start_line);
		}

"<"[a-zA-Z0-9_]*">"	{
		  int c = input (), state = -1;
		  tom_int start_line = ilx->current_line;
		  char *text = alloca (1 + yyleng);
		
		  /* Copy yytext, as that pointer can be mangled by
                     input().  */
		  memcpy (text, yytext, yyleng);
		  while (c != EOF)
		    {
		      if (c == '<')
			state = 0;
		      else if (state == 0)
			state = c == '/' ? state + 1 : -1;
		      else if (state > 0)
			if (text[state] == c)
			  {
			    if (c == '>')
			      break;
			    state++;
			  }
			else
			  state = -1;
		
		      if (c == '\n')
			ilx->current_line++;
		
		      c = input ();
		    }
		
		  if (c == EOF)
		    error ("%d: unterminated %s documentation",
			   start_line, text);
		}

[ \t\f]*\n	{ ilx->current_line++; }
[ \t\f]*	{ /* Retry.  */ }

[A-Za-z][-A-Za-z_0-9]*[A-Za-z_0-9]|[A-Za-z][A-Za-z_0-9]* {
		  ilx->semantics
		    = (void *) byte_string_with_string
		    (c_string_with_length (yytext, yyleng), yyleng);
		  return ilx->token = c_gp_GPParseDefines_IDENTIFIER;
		}

[0-9]+		{
		  char *buf = yytext, *next = yytext + yyleng;
		  tom_int v, d;
		  char *n;

		  /* Maybe itl'll fit an int.  */
		  for (v = 0, n = buf; n < next; n++)
		    {
		      d = *n - '0';
		      if ((INT_MAX - d) / 10 >= v)
			v = v * 10 + d;
		      else
			break;
		    }
		
		  if (n == next)
		    ilx->semantics
		      = (void *) TRT_SEND ((reference_imp),
					   _mr_c_tom_IntNumber,
					   USEL (tom, r_with_i), v);
		  else
		    {
		      /* Hopefully, it'll fit a long long.  */
		      tom_long lv, nlv;
		
		      for (lv = 0, n = buf; n < next; n++)
			{
			  d = *n - '0';
			  nlv = lv * 10 + d;
			  if ((nlv - d) / 10 != lv)
			    {
			      *next = 0;
			      error ("number too large: %s", buf);
			    }
			  lv = nlv;
			}
		      ilx->semantics = TRT_SEND (_PI_, _mr_c_tom_LongNumber,
						 USEL (tom, r_with_l), lv);
		    }
		  return ilx->token = c_gp_GPParseDefines_NUMBER;
		}

"{"		{
		  read_compound (ilx, '{', '}');
		  return ilx->token = c_gp_GPParseDefines_EXPR;
		}

"%%"		{
		  read_flat_compound (ilx, "%%");
		  return ilx->token = c_gp_GPParseDefines_CODE;
		}

'(\\.|[^\\'])*' {
		  tom_object s = lex_string_cst (ilx, yytext, yyleng, '\'');
		  tom_byte b;

		  if (TRT_SEND (, s, SEL (i_length)) != 1)
		    {
		      error ("bad ASCII byte constant");
		      b = 0;
		    }
		  else
		    b = TRT_SEND (, s, SEL (b_at_i), 0);

		  /* semantics = [parser terminal b];  */
		  ilx->semantics
		    = TRT_SEND (_PI_, ilx->parser, USEL (gp, r_terminal_i),
				(tom_int) b);

		  return ilx->token = c_gp_GPParseDefines_LITERAL;
		}

"%token"	{ return ilx->token = c_gp_GPParseDefines_TOKEN; }
"%class"	{ return ilx->token = c_gp_GPParseDefines_CLASS; }
"%instance"	{ return ilx->token = c_gp_GPParseDefines_INSTANCE; }
"%predef"	{ return ilx->token = c_gp_GPParseDefines_PREDEF; }

.		{ return ilx->token = yytext[0]; }

%%
static void
read_flat_compound (struct _es_i_gps_SkeletonLexer *ilx, char *close)
{
  tom_int start_line = ilx->current_line;
  int next_match = 0, i, c = input ();
  tom_object s;

  s = TRT_SEND ((reference_imp), _mr_c_tom_MutableByteString, SEL (r_new));

  /* XXX This does not handle comments, strings, documentation, etc.  */
  while (c != EOF)
    if (c == close[next_match])
      {
	if (!close[++next_match])
	  break;
	c = input ();
      }
    else
      {
	/* XXX Ehhh, backtrack.  */
	if (next_match)
	  {
	    for (i = 0; i < next_match; i++)
	      TRT_SEND ((void_imp), s, SEL (v_add_b), (tom_byte) close[i]);
	    next_match = 0;
	  }

	TRT_SEND ((void_imp), s, SEL (v_add_b), (tom_byte) c);

	if (c == '\n')
	  ilx->current_line++;
	c = input ();
      }

  if (c == EOF)
    error ("%d: unterminated flat compound", start_line);

  ilx->semantics = (void *) s;
}

static void
read_compound (struct _es_i_gps_SkeletonLexer *ilx, char open, char close)
{
  tom_int start_line = ilx->current_line;
  int nest_level = 0, c = input ();
  tom_object s;

  enum
  {
    S_TOP,
    S_STRING,
    S_STRING_ESC,
    S_CHAR,
    S_CHAR_ESC,
    S_SLASH,
    S_CPP_COMMENT,
    S_C_COMMENT,
    S_C_COMMENT_STAR,
  } state = S_TOP;

  s = TRT_SEND ((reference_imp), _mr_c_tom_MutableByteString, SEL (r_new));

  /* XXX This does not handle documentation.
     Wed Apr  9 12:17:37 1997, tiggr@cobra.ics.ele.tue.nl  */
  while (c != EOF && (c != '}' || nest_level || state != S_TOP))
    {
      TRT_SEND ((void_imp), s, SEL (v_add_b), (tom_byte) c);

      if (c == '\n')
	{
	  ilx->current_line++;
	  if (state == S_CPP_COMMENT)
	    state = S_TOP;
	}

      if (state == S_SLASH)
	{
	  if (c == '*')
	    {
	      /* We don't employ enough states to discern
                 `slash-star-slash' as not being a comment.  */
	      state = S_C_COMMENT;
	    }
	  else if (c == '/')
	    state = S_CPP_COMMENT;
	  else
	    state = S_TOP;
	}

      if (state == S_TOP)
	{
	  if (c == '{' || c == '(')
	    nest_level++;
	  else if (c == '}' || c == ')')
	    nest_level--;
	  else if (c == '"')
	    state = S_STRING;
	  else if (c == '\'')
	    state = S_CHAR;
	  else if (c == '/')
	    state = S_SLASH;
	}
      else if (state == S_STRING)
	{
	  if (c == '\\')
	    state = S_STRING_ESC;
	  else if (c == '"')
	    state = S_TOP;
	}
      else if (state == S_STRING_ESC)
	state = S_STRING;
      else if (state == S_CHAR)
	{
	  if (c == '\\')
	    state = S_CHAR_ESC;
	  else if (c == '\'')
	    state = S_TOP;
	}
      else if (state == S_CHAR_ESC)
	state = S_CHAR;
      else if (state == S_C_COMMENT)
	{
	  if (c == '*')
	    state = S_C_COMMENT_STAR;
	}
      else if (state == S_C_COMMENT_STAR)
	{
	  if (c == '*')
	    ;
	  else if (c == '/')
	    state = S_TOP;
	}

      c = input ();
    }
  if (c == EOF)
    error ("%d: unterminated compound", start_line);

  ilx->semantics = (void *) s;
}
